from arcalg import loops, BasisElement, Element, E0
from collections import UserDict
import numpy as np

try:
    import galois
    galAvailable = True
except:
    galAvailable = False

try:
    import snappy
    snappyAvailable = True
except:
    snappyAvailable = False


class ChainCx:
    """ chain complex class """
    """ contains the data:
        - list of generators
        - for each generator xless matching (i.e. associated idempotent)
            (or a basis element, if the input has been capped.)
        - differential, maps each generator to an Element
        - capped, which is a boolean noting whether the 
            input has been capped.
        - gr - dictionary with keys generators and values relative homological and quantum gradings.
        - tangle - a list of the cups, caps, and crossings that created this complex.
    """
    def __init__(self,tangle, simplify=True): 
        """ inputs: a tangle represented as a list of 
                elementary tangles and two crossingless matchings
                (xlm etangles xlm)
             simplify indicates whether to simplify after each gluing step; True is faster.
        """
        # tangle conventions: input is from right to left. as one would read it.
        self.capped = False
        self.tangle = list()
        if len(tangle) == 1:
            self.generators = {'0':tangle[0]} 
            self.differential = {'0':{}} 
            self.gr = {'0':(0, 0)} #Stores the *relative* (homological,quantum) grading, i.e., the correct gradings up to an overall shift.
            self.tangle.append(tangle)
        elif len(tangle) > 1:
            self.tangle = list(tangle)
            # adds the first crossingless matching
            tempCx = ChainCx([tangle[-1]])
            endpoint = len(tangle)
            
            #adds the elementary tangles
            if len(tangle[0]) % 2 == 1:
                endpoint += 1
            for i in range(2,endpoint):
                tempCx.glue(tangle[-i])
                if simplify:
                    simp, gen, image = tempCx.simplifiable()
                    while simp:
                        tempCx.simplify(gen, image)
                        simp, gen, image = tempCx.simplifiable()
            
            # if the last one is also a crossingless matching we add it separately.
            if len(tangle[0]) % 2 == 0:
                tempCx.attachCaps(tangle[0]) # attach all the caps!
                tempCx.capped = True #modifies simplify/able 

            # one last simplification
            if simplify:
                simp, gen, image = tempCx.simplifiable()
                while simp:
                    tempCx.simplify(gen, image)
                    simp, gen, image = tempCx.simplifiable()
            
            self.generators = tempCx.generators
            self.differential = tempCx.differential
            self.gr = tempCx.gr
            self.capped = tempCx.capped

        else: 
            # in case we want some other constructor perhaps?
            pass
        
    def __repr__(self):
        s = ''
        for gen in self.generators:
            s = s + str(gen) + " with value: "+ str(self.generators[gen]) 
            s = s + " has diff: " +  str(self.differential[gen]) + '\n'
        return s

    def glue(self, etangle):
        """ modifies the chain complex self by 
            gluing on the given elementary tangle.
            creates new generators, calculates 
            the new differential, and renames 
            all generators.
            may introduce summands that 
            are chain homotopy equivalent 
            to 0.
        """
        if self.capped:
            raise Exception("can't glue onto capped complex")
        ox = ''
        unox = ''
        self.tangle.append(etangle)
        if etangle[0] == 'u': #up case
            ox = 'z'
            unox = 'w'
        elif etangle[0] == 'd': # down case
            ox = 'w'
            unox = 'z'
        elif etangle[0] == 'cup': # cup case!
            newgens, newdiff,  newgr = glueCup(etangle[1],self.generators,self.differential, self.gr)
            self.generators, self.differential,  self.gr = rename(newgens, newdiff,  newgr)
            return
        else: #cap case
            raise(NotImplementedError)
            #newgens, newdiff, newgr = glueCap(etangle[1],self.generators, self.differential,self.gr)
            return 
        newgens, newdiff,  newgr = glueUD(etangle[1], ox, unox, self.generators, self.differential,  self.gr)
        self.generators, self.differential,  self.gr = rename(newgens, newdiff,  newgr)
        return
        
    def simplify(self, gen=None, image=None):
        """ finds summands of self that 
            are chain homotopy equivalent to 
            zero and removes them.
            if there are none, nothing happens.
            if we give it a gen it assumes that 
            that has a differential acting by one.
        """
        if gen == None:
            simp, gen, image = self.simplifiable()
            if not simp:
                return # we just don't do anything


        # preliminary tidying!
        self.generators.pop(gen)
        self.generators.pop(image)
        self.gr.pop(gen)
        self.gr.pop(image)

        # setting up + etc.
        self.differential[gen].pop(image)
        remainder = self.differential[gen]
        if self.capped:
            for i in remainder:
                remainder[i] *= -1
        else:
            for i in remainder:
                remainder[i].scalarMult(-1)
        self.differential.pop(gen)
        self.differential.pop(image)


        for generator in self.generators:
            if gen in self.differential[generator]:
                # remove it; it doesn't matter
                self.differential[generator].pop(gen)
            if image not in self.differential[generator]:
                # coefficient is zero! we proceed unchanged.
                continue
            coeff = self.differential[generator].pop(image)
            for imgen in remainder:
                if imgen not in self.differential[generator]: #If this wasn't a term in d(generator), add it.
                    if self.capped:
                        self.differential[generator][imgen] = remainder[imgen]*coeff
                    else:
                        self.differential[generator][imgen] = coeff.mult(remainder[imgen])
                else: #If this was a term in d(generator), add it.
                    if self.capped:
                        addend = remainder[imgen]*coeff
                        self.differential[generator][imgen] += addend
                    else:
                        # a note: self.differential[generator][imgen] is an Element.
                        addend = coeff.mult(remainder[imgen])
                        self.differential[generator][imgen] = self.differential[generator][imgen].add(addend)
    
        return
    
    def simplifiable(self):
        """ returns true + generator if self has a summand 
            chain homotopy equivalent to zero, 
            otherwise returns false (and None for 
            niceness)
        """
        if self.capped:
            for gen, diff in self.differential.items():
                for image in diff:
                    if diff[image] == 1:
                        return True, gen, image
            return False, None, None

        # finds arrows w coeff 1 that we can remove
        for gen, diff in self.differential.items():
            for image in diff:
                idem = self.generators[image]
                if diff[image] == BasisElement(idem,idem,loops(idem,idem)):
                    # the idempotent associated with image
                    # WITH ones everywhere.
                    return True, gen, image

        return False, None, None

    
    def attachCaps(self, xlm):
        """ attaches caps corresponding to the 
            given crossingless matching.

            ALSO: doesn't respect base 2 stuff!!
            (even if we're using mods)
        """
        self.capped = True
        self.tangle = [xlm] + self.tangle
        # create all our helper dicts!
        newgens = {} # keys: new gens (string names again), values: BasisElements (NOT idems.)
        gendict = {} # keys: old gens, values: new gens
        newgr = {}
            
        # create the new generators!
        # what we're doing:
        # for each old gen, we attach the given xlm 
        #    on the left w old idem on the right
        # then we give it All possible labellings.
        for oldgen, idem in self.generators.items():
            circles = loops(xlm, idem)
            for k in range(2**len(circles)):
                name = str(k) + 'g' + oldgen
                gen = BasisElement(xlm, idem, circles, k)
                newgens[name] = gen
                newgr[name] = (self.gr[oldgen][0],self.gr[oldgen][1]+2*_base_2_size(k)-len(circles))
                if oldgen in gendict:
                    gendict[oldgen] += [name]
                else:
                    gendict[oldgen] = [name]

        newdiff = {}
        # create differentials!
        # for each new gen, multiply w differential from its 
        #     old gen. the result is the map!. note that 
        #     coeffs are in fact just integers.
        for oldgen, diff in self.differential.items():
            for gen in gendict[oldgen]:
                newdiff[gen] = {}
                for diffgen, elt in diff.items():
                    result = Element(newgens[gen]).mult(elt)
                    for basiselt, coeff in result.dict.items():
                        if coeff == 0:
                            continue
                        k = basiselt.getK()
                        newdg = str(k) + "g" + diffgen
                        if newdg in newdiff[gen]:
                            newdiff[gen][newdg] += coeff
                        else:
                            newdiff[gen][newdg] = coeff

        self.generators, self.differential,  self.gr = rename(newgens, newdiff,  newgr)
        return

    def test_d_squared(self):
        "Return True if differential squares to zero. Otherwise, returns false and prints an example of the failure."
        trueorfalse = True
        for g in self.generators.keys():
            dg = self.differential[g]
            answer = ME0
            for h in dg.keys():
                answer += dg[h]*ModuleElement(self.differential[h])
            if not self.capped:#Remove zeroes.
                for x in list(answer.keys()):
                    for y in list(answer[x].dict.keys()):
                        if answer[x].dict[y] == 0: #An algebra element with coefficient 0 (in the integers)
                            answer[x].dict.pop(y)
                    if answer[x].dict == {}: #The zero algebra element
                        answer.pop(x)
            else:
                for x in list(answer.keys()): 
                    if answer[x] == 0:
                        answer.pop(x)
            if answer:
                print("Failed d^2=0: d^2("+repr(g)+")="+repr(answer))
                trueorfalse = False
        return trueorfalse

    def test_diff_gr(self):
        "Tests whether the differential has bidegree (-1,0)."
        answer = True
        for x in self.generators:
            for y in self.differential[x]:
                if self.gr[y][0] != self.gr[x][0]-1: #Check the homological gradings.
                    print("Generator "+repr(x)+" has grading "+repr(self.gr[x])+" but differential contains "+repr(y)+" with homological grading "+repr(self.gr[y][0]))
                    answer = False
                if self.capped: #Quantum gradingds, closed case
                    if self.gr[y][1] != self.gr[x][1]: #Quantum gradings disagree.
                        print("Generator "+repr(x)+" has grading "+repr(self.gr[x])+" but differential contains "+repr(y)+" with quantum grading "+repr(self.gr[y][1]))
                        answer = False
                else: #Quantum gradings, non-closed case
                    a = self.differential[x][y]
                    for (b, c) in a.dict.items():
                        if c:
                            grb = 2*sum(b.labels)-len(b.loops)+int(len(b.cm1)/2)
                            if self.gr[x][1] != self.gr[y][1] + grb: #Do quantum gradings agree?
                                print("Generator "+repr(x)+" has grading "+repr(self.gr[x])+" but differential contains "+repr(y)+" with quantum grading "+repr(self.gr[y][1] + grb))
                                answer = False
        return answer

    def rational_homology(self):
        """
        Returns the total dimension of the (Khovanov) homology of this link with rational coefficients. Only works for (closed) links, not tangles.
        If SnapPy is available, returns absolutely graded complex. Otherwise, result has correct relative gradings, but not correct absolute gradings.
        """
        assert self.capped #Only makes sense for closed links, not tangles.
        n = len(self.generators.keys()) #number of generators
        diffmx = np.array(n*[n*[0, ]]) #square array of zeroes
        genlist = list(self.generators.keys())
        name_to_pos = dict()
        for (i, x) in enumerate(genlist):
            name_to_pos[x]=i
        for i in range(n):
            x = genlist[i]
            dx = self.differential[x]
            for (y, a) in dx.items():
                diffmx[i, name_to_pos[y]] += a
        assert (np.matmul(diffmx, diffmx) == 0).all() #Check that d^2=0, in obnoxious numpy syntax.
        rank = np.linalg.matrix_rank(diffmx)
        return n-2*rank

    def graded_rat_homol(self):
        "Returns the relatively bigraded rational (Khovanov) homology of this link. Only works for closed links, not tangles."
        assert self.capped #Only makes sense for closed links, not tangles.
        gens_by_grading = dict() #Dictionary with keys bigradings and values generators in those bigradings.
        newgr = dict()
        if snappyAvailable:
            nplus = self.nplus()
            nminus = self.nminus()
            for x in self.generators.keys():
                newgr[x] = (self.gr[x][0]+nminus,self.gr[x][1]-nplus+2*nminus) #Add correction here.
        else:
            newgr = self.gr
        for x in self.generators.keys():
            try:
                gens_by_grading[newgr[x]].append(x)
            except:
                gens_by_grading[newgr[x]] = [x,]
        answer = dict() #Dictionary with keys bigradings and values the dimension of Khovanov homology in that bigrading
        for (i,j) in newgr.values():
            try:
                m = len(gens_by_grading[(i+1,j)]) #Number of generators one grading up.
            except:
                m=0
                gens_by_grading[(i+1,j)] = list()
            n = len(gens_by_grading[(i,j)]) #number of generators in this grading
            try:
                p = len(gens_by_grading[(i-1,j)]) #Number of generators one grading down
            except:
                p=0
                gens_by_grading[(i-1,j)] = list()
            #npdiffmx = np.array(p*[n*[0, ]]) #Create matrix *out* of grading (i,j); start out all zeroes
            npdiffmx = np.array(n*[p*[0, ]]) #Create matrix *out* of grading (i,j); start out all zeroes
            mndiffmx = np.array(m*[n*[0, ]]) #Create matrix *into* grading (i,j); start out all zeroes
            name_to_pos = dict()
            for (k, x) in enumerate(gens_by_grading[(i,j)]):
                name_to_pos[x]=k
            for (k, x) in enumerate(gens_by_grading[(i+1,j)]):
                name_to_pos[x]=k
            for (k, x) in enumerate(gens_by_grading[(i-1,j)]):
                name_to_pos[x]=k
            for k in range(n):
                x = gens_by_grading[(i,j)][k]
                dx = self.differential[x]
                for (y, a) in dx.items():
                    npdiffmx[k, name_to_pos[y]] += a
            for k in range(m):
                x = gens_by_grading[(i+1,j)][k]
                dx = self.differential[x]
                for (y, a) in dx.items():
                    mndiffmx[k, name_to_pos[y]] += a
            if m>0 and p>0:
                assert (np.matmul(mndiffmx, npdiffmx) == 0).all() #Check that d^2=0, in obnoxious numpy syntax.
            if m > 0:
                rank = np.linalg.matrix_rank(mndiffmx)
            else:
                rank = 0
            if p>0:
                nullity = n-np.linalg.matrix_rank(npdiffmx)
            else:
                nullity = n
            assert nullity >= rank
            answer[(i,j)] = nullity-rank
        return answer

    def graded_Fp_homol(self,prime):
        """Returns the bigraded (Khovanov) homology of this link with coefficients in the field or order p. Only works for closed links, not tangles.
        If SnapPy is available, returns absolutely graded complex. Otherwise, result has correct relative gradings, but not correct absolute gradings.
        """
        assert self.capped #Only makes sense for closed links, not tangles.
        if not galAvailable:
            raise Exception("Requires Python package galois, which seems not to be available.")
        GF = galois.GF(prime)
        if snappyAvailable:
            newgr = dict()
            nplus = self.nplus()
            nminus = self.nminus()
            for x in self.generators.keys():
                newgr[x] = (self.gr[x][0]+nminus,self.gr[x][1]-nplus+2*nminus) #Add correction here.
        else:
            newgr = self.gr
        gens_by_grading = dict() #Dictionary with keys bigradings and values generators in those bigradings.
        for x in self.generators.keys():
            try:
                gens_by_grading[newgr[x]].append(x)
            except:
                gens_by_grading[newgr[x]] = [x,]
        answer = dict() #Dictionary with keys bigradings and values the dimension of Khovanov homology in that bigrading
        for (i,j) in newgr.values():
            try:
                m = len(gens_by_grading[(i+1,j)]) #Number of generators one grading up.
            except:
                m=0
                gens_by_grading[(i+1,j)] = list()
            n = len(gens_by_grading[(i,j)]) #number of generators in this grading
            try:
                p = len(gens_by_grading[(i-1,j)]) #Number of generators one grading down
            except:
                p=0
                gens_by_grading[(i-1,j)] = list()
            npdiffmx = np.array(n*[p*[0, ]],dtype=int) #Create matrix *out* of grading (i,j); start out all zeroes
            mndiffmx = np.array(m*[n*[0, ]],dtype=int) #Create matrix *into* grading (i,j); start out all zeroes
            name_to_pos = dict()
            for (k, x) in enumerate(gens_by_grading[(i,j)]):
                name_to_pos[x]=k
            for (k, x) in enumerate(gens_by_grading[(i+1,j)]):
                name_to_pos[x]=k
            for (k, x) in enumerate(gens_by_grading[(i-1,j)]):
                name_to_pos[x]=k
            for k in range(n):
                x = gens_by_grading[(i,j)][k]
                dx = self.differential[x]
                for (y, a) in dx.items():
                    npdiffmx[k, name_to_pos[y]] += a
            for k in range(m):
                x = gens_by_grading[(i+1,j)][k]
                dx = self.differential[x]
                for (y, a) in dx.items():
                    mndiffmx[k, name_to_pos[y]] += a
            mndiffmx = GF(mndiffmx % prime)
            npdiffmx = GF(npdiffmx % prime)
            if m>0 and p>0:
                assert (np.matmul(mndiffmx, npdiffmx) == 0).all() #Check that d^2=0, in obnoxious numpy syntax.                
            if m > 0:
                rank = np.linalg.matrix_rank(mndiffmx)
            else:
                rank = 0
            if p>0:
                nullity = n-np.linalg.matrix_rank(npdiffmx)
            else:
                nullity = n
            assert nullity >= rank
            answer[(i,j)] = nullity-rank
        return answer

    def to_snappy(self):
        "Returns a SnapPy MorseEncoding of self. To get a SnapPy link, call its .link() function. Assumes tangle has the form (cups, crossings, caps), i.e., that there are no cups or caps in the middle."
        if not self.capped:
            raise Exception("Can only export closed links, not tangles, to SnapPy")
        if not snappyAvailable:
            raise Exception("SnapPy required but seems not to be available")
        from spherogram.links.exhaust import MorseEncoding
        cm1 = self.tangle[0]
        cm2 = self.tangle[-1]
        events = list()
        #The cups from cm1
        not_used = list(range(len(cm1)))
        for i in range(len(cm1)):
            if i in not_used:
                events.append(('cup',i,i+1))
                not_used.remove(i)
                not_used.remove(cm1[i])
        #The crossings
        for c in self.tangle[1:-1]:
            if c[0] == 'u':
                events.append(('cross',c[1],c[1]+1))
            else:
                events.append(('cross',c[1]+1,c[1]))
        #The caps from cm2
        caps = list()
        not_used = list(range(len(cm1)))
        for i in range(len(cm2)):
            if i in not_used:
                caps.append(('cap',i,i+1))
                not_used.remove(i)
                not_used.remove(cm2[i])
        caps.reverse()
        events.extend(caps)
        return MorseEncoding(events)

    def nplus(self):
        "Returns number of positive crossings of self"
        K = self.to_snappy().link()
        return sum([c.sign for c in K.crossings if c.sign == 1])

    def nminus(self):
        "Returns number of positive crossings of self"
        K = self.to_snappy().link()
        return abs(sum([c.sign for c in K.crossings if c.sign == -1]))

# helper functions!!

def rename(gens, diff,  grs):
    """ inputs: gens, a list of generators,
        diff, a dictionary giving all differentials.

        renames generators and updates 
        the differential to match!
        so that we don't have superexcessive 
        names walking around (and so that 
        my glue function works as expected...)
    """
    nameLen = len(str(len(gens)))
    trans = {}
    newgens = {}
    newdiff = {}
    newgr = {}

    # make new names and correspond them.
    for i, gen in zip(range(len(gens)), gens):
        name = (nameLen-len(str(i)))*'0' + str(i)
        trans[gen] = name
    
    # update generators
    for gen in gens:
        newgens[trans[gen]] = gens[gen]
        newgr[trans[gen]] = grs[gen]
    
    # update differential (annoyingest deep copy ever.)
    for gen in diff:
        newdiff[trans[gen]] = {}
        for imgen in diff[gen]:
            newdiff[trans[gen]][trans[imgen]] = diff[gen][imgen]
    
    return newgens, newdiff,  newgr

def attach(index, xlm):
    """ takes in an index and a crossingless 
        matching. 
        returns a new crossingless matching and 
        an indicator of whether the new crossingless 
        matching also contains an extra circle.
        (index represents where cup/cap smoothing 
        from a crossing begins)
    """
    if xlm[index] == index + 1:
        return xlm, True 
    else:
        out = [0]*len(xlm)
        for i in (range(len(xlm))):
            if i == index:
                out[i] = index + 1
            elif i == index + 1:
                out[i] = index
            elif xlm[i] not in [index,index+1]:
                out[i] = xlm[i]
            elif xlm[i] == index:
                out[i] = xlm[index+1]
            else:
                out[i] = xlm[index]
        return tuple(out), False

def attachcup(index, xlm):
    """ attaches a cup to given crossingless 
        matching. never creates an extra circle.
    """
    out = [0]*(len(xlm)+2)
    for i in range(len(out)):
        if i < index:
            if xlm[i] < index:
                out[i] = xlm[i]
            else:
                out[i] = xlm[i] + 2
        elif i == index:
            out[i] = index+1
        elif i == index + 1:
            out[i] = index 
        else:
            if xlm[i-2] < index:
                out[i] = xlm[i-2]
            else:
                out[i] = xlm[i-2] + 2
    return tuple(out)

def attachcap(index,xlm):
    """ attaches at cap at the given index 
        to a crossingless matching. might create
        an extra circle!
    """
    prelimout, extra = attach(index,xlm)
    out = [0]*(len(xlm)-2)
    for i in range(len(out)):
        if i < index:
            if prelimout[i] < index:
                out[i] = prelimout[i]
            else:
                out[i] = prelimout[i] - 2
        else:
            if prelimout[i+2] < index:
                out[i] = prelimout[i+2]
            else:
                out[i] = prelimout[i+2] - 2
    return tuple(out), extra

def refactor(etangle, basiselt, extraIsX=True):
    """ turns gen/alg element product 
        to alg elt/gen product in the 
        bimodule associated with etangle
        (just doing a basis element 
        because we can very easily loop
        through an element to just get 
        this case.)
    """"""
        (WHEN alphaprime etangle alpha has 
        an extra circle labeled x.. *or not*)
        (hence alphaprime = alpha.)
    """
    alpha = basiselt.cm1
    beta = basiselt.cm2 
    origLoops = basiselt.loops
    origLabels = basiselt.labels
    outExtra = None

    index = etangle[1]

    if etangle[0] in 'ud':
        betaprime, match = attach(index,beta)
    elif etangle[0] == 'cap':
        betaprime, match = attachcap(index,beta)
        # i will probably need to be FAR more careful with my indices in this case??.......
    else: # i'm not sure this else case will ever happen???
        betaprime = attachcup(etangle[1],beta)
        match = False 
    
    if not match:
        alphaprime = attach(index,alpha)[0]
    else:
        alphaprime = alpha

    circles = loops(alphaprime,betaprime)

    for i in range(len(origLoops)):
        if index in origLoops[i]:
            if extraIsX and origLabels[i] == 1:
                # if the circle that gets merged with 
                # our extra little circle is labelled x....
                return 0, betaprime, None # and something???

    if match:
        outExtra = 0
        labels = list(origLabels)
        for i in range(len(circles)):
            if extraIsX and index in circles[i]:
                # gotta label that little one
                labels[i] = 1
                break # for runtime reasons but doesn't do much..
        for i in range(len(circles)):
            # i suspect there is a more clever way to do this..
            if index in circles[i] and labels[i] == 1:
                labels[i] = 0
                outExtra = 1
                break # same runtime reasons.
            
    else:
        labels = [0]*len(circles)
        for i in range(len(circles)):
            if index in circles[i]:
                # circle labelled 1, so we can carry on
                continue 
            elif extraIsX and beta[index] in circles[i]:
                labels[i] = 1
            else:
                for j in range(len(origLoops)):
                    #i'm sorry @ my runtime...
                    if not set(circles[i]).isdisjoint(origLoops[j]) and labels[i] != 1:
                        labels[i] = origLabels[j]
                        

    return BasisElement(alphaprime,betaprime,circles,inlabels=labels), betaprime, outExtra

def refactor_split(index, basiselt, xlm1):
    """ like refactor but does split things 
        for when we're mapping from no circle
        to circle w 1 and circle w x.
    """"""
        i have NOO idea if this works even at all.
    """
    alpha = basiselt.cm1
    beta = basiselt.cm2 
    origLoops = basiselt.loops
    origLabels = basiselt.labels

    betaprime, match = attach(index,beta)
    # match SHOULD be true so idk why we create it.

    circles = loops(xlm1,betaprime)    

    labels1 = [0]*len(circles)
    labels2 = [0]*len(circles)
    for i in range(len(circles)):
        if alpha[index] in circles[i]:
            labels1[i] = 1
        elif index in circles[i]:
            continue
        else:
            rep = circles[i][0]
            for j in range(len(origLoops)):
                if rep in origLoops[j]:
                    labels1[i] = origLabels[j]
                    labels2[i] = origLabels[j]

    for i in range(len(origLoops)):
        if index in origLoops[i]:
            if origLabels[i] == 1:
                # if the circle that gets split is labelled x: 
                # the splitting is just x and x!!
                return 0, BasisElement(alpha,betaprime,circles,inlabels=labels1)
    
    return BasisElement(xlm1,betaprime,circles,inlabels=labels1), \
        BasisElement(xlm1,betaprime,circles,inlabels=labels2)

def refactor_cap(index, basiselt, xlm1):
    """ refactor but when there's a cap at index.
        basiselt is the thing being refactored 
        against the idem associated w xlm1.
    """

    return 

def glueUD(index, ox, unox, generators, differential, grs):
    """ the helper function for glue when 
        the elementary tangle is either up or down.
        ox is the smoothing that is easy 
        and unox is the smoothing that is hard
    """
    # instantiate all our helpers :)
    newgens = {} # the new dictionary that is the generators (names:idems)
    gendict = {} # keys: old gens, values: new gens
    smoothdict = {} # keys: zero smoothings (new gens), values: corr 1-smoothings
    idemdict = {} # keys: idempotents, values: new gens w that idempotent.
    newgr = {} #keys new generators, values their (h,q) bigrading

    if ox == 'z': #up case
        hgrcor = {ox:0, unox:1}
        qgrcor = {ox:0, unox:1}
    else: #down case
        hgrcor = {ox:1, unox:0}
        qgrcor = {ox:1, unox:0}

    # create new generators!
    for gen, idem in generators.items():
        # ox smoothing (id braid)
        esmth = ox + gen
        newgens[esmth] = idem
        gendict[gen] = [esmth]
        (oldh, oldq) = grs[gen]
        if idem in idemdict:
            idemdict[idem] += [esmth]
        else:
            idemdict[idem] = [esmth]
        newgr[ox+gen] = (oldh-hgrcor[ox], oldq-qgrcor[ox]) 

        # unox smoothing (loops)
        newidem, extra = attach(index, idem) # calculate new crossingless matching.
        if extra: # there's an extra circle
            hsmth = [unox + 'a' + gen, unox + 'x' + gen]
            newgr[unox+'a'+gen] = (oldh-hgrcor[unox], oldq-1-qgrcor[unox])
            newgr[unox+'x'+gen] = (oldh-hgrcor[unox], oldq+1-qgrcor[unox]) #RL: have set quantum grading of X to be 2, of 1 to be 0.
        else: # there's not
            hsmth = [unox + gen]
            newgr[unox+gen] = (oldh-hgrcor[unox], oldq-qgrcor[unox])
        for smth in hsmth:
            newgens[smth] = newidem
    
        gendict[gen] += hsmth
        if ox == 'z': #if it's up
            smoothdict[esmth] = hsmth.copy()
        else: #otherwise if it's down.
            for smth in hsmth:
                smoothdict[smth] = [esmth]
        if newidem in idemdict:
            idemdict[newidem] += hsmth
        else:
            idemdict[newidem] = hsmth.copy()
    
    # differential! 
    newdiff = {}

    # terms from old differential
    for oldgen, diff in differential.items():
        # diff is a dictionary {oldgens:Elements}
        for gen in gendict[oldgen]:
            newdiff[gen] = {}
            if ox in gen:
            # the everything's nice case :)
                koszul = (-1)**(hgrcor[ox])
                for diffgen in diff:
                    newdiff[gen][ox+diffgen] = diff[diffgen]*koszul
            elif 'x' not in gen and 'a' not in gen:
                koszul = (-1)**(hgrcor[unox])
                #no circle
                for diffgen in diff:
                    if len(gendict[diffgen]) == 2:
                        # no circles in new gens
                        newelt = Element(dictionary = {})
                        oldelt = diff[diffgen]
                        xlm1 = newgens[gen] #idem of generator we are finding diff of.
                        for basiselt in oldelt.dict:
                            newBE = refactor(('u',index,len(xlm1)),basiselt,False)[0]
                            newelt = newelt.add(Element(newBE,oldelt.dict[basiselt]))
                        newdiff[gen][unox+diffgen] = newelt*koszul
                    else:
                        newelt1 = Element(dictionary={})
                        newelt2 = Element(dictionary={})
                        oldelt = diff[diffgen]
                        xlm1 = newgens[gen]
                        for basiselt in oldelt.dict:
                            nbe1, nbe2 = refactor_split(index, basiselt, xlm1)
                            if nbe1 != 0: # if it's zero there is no contribution.
                                newelt1.addInPlace(Element(nbe1, oldelt.dict[basiselt]))
                            newelt2.addInPlace(Element(nbe2, oldelt.dict[basiselt]))
                        newdiff[gen][unox + 'a' + diffgen] = newelt1*koszul
                        newdiff[gen][unox + 'x' + diffgen] = newelt2*koszul
            else:
                koszul = (-1)**(hgrcor[unox])
                for diffgen in diff:
                    if len(gendict[diffgen]) == 3:
                        # circles in new gens!
                        # this is a little broken actually.
                        newelt0 = Element(dictionary = {})
                        newelt1 = Element(dictionary = {})
                        oldelt = diff[diffgen]
                        etangle = ('u',index,len(newgens[gen]))
                        # it doesn't actually matter whether it's u or d for refactor.
                        extra = 'x' in gen
                        for basiselt, coeff in oldelt.dict.items():
                            newBE, idem, label = refactor(etangle, basiselt, extra)
                            if newBE == 0:
                                continue
                            if label == 0:
                                newelt0 = newelt0.add(Element(newBE,coeff))
                            elif label == 1:
                                newelt1 = newelt1.add(Element(newBE,coeff))
                            # no else case bc i don't expect there to be one.
                        newdiff[gen][gen[0]+'a'+diffgen] = newelt0*koszul
                        newdiff[gen][gen[0]+'x'+diffgen] = newelt1*koszul
                    else:
                        # there are MERGES.
                        extraX = 'x' in gen
                        newelt = Element(dictionary = {})
                        oldelt = diff[diffgen]
                        xlm1 = newgens[gen]
                        for basiselt in oldelt.dict:
                            newdiffgen, xlm2, label = refactor(('d', index, len(xlm1)), basiselt, extraX)
                            newelt = newelt.add(Element(newdiffgen, oldelt.dict[basiselt]))
                        newdiff[gen][unox+diffgen] = newelt*koszul
            
    # terms from new differential
    for gen, idem in newgens.items():
        if 'w' in gen: # going for 0-smoothings here!
            continue
        imgen = smoothdict[gen][0]
        imidem = newgens[imgen]
        circles = loops(idem,imidem)
        coeff = BasisElement(idem,imidem,circles)
        Ecoeff = Element(coeff)
        if 'x' in gen: # down case, extra circle, labelled x
            cic = 0
            for i in range(len(circles)):
                if index in circles[i]:
                    cic = len(circles)-i-1
                    break
            Ncoeff = Element(BasisElement(idem,imidem,circles,2**cic)) 
            if gen in newdiff:
                if imgen in newdiff[gen]:
                    newdiff[gen][imgen] = newdiff[gen][imgen].add(Ncoeff)
                else:
                    newdiff[gen][imgen] = Ncoeff
            else:
                newdiff[gen] = {imgen : Ncoeff} 
        elif ox == 'w': #down case, rest of cases
            if gen in newdiff:
                if imgen in newdiff[gen]:
                    newdiff[gen][imgen] = newdiff[gen][imgen].add(Ecoeff)
                else:
                    newdiff[gen][imgen] = Ecoeff
            else:
                newdiff[gen] = {imgen : Ecoeff}

        #up cases:
        elif len(smoothdict[gen]) == 1: # no extra circles in image!!!
            if gen in newdiff:
                if imgen in newdiff[gen]:
                    newdiff[gen][imgen].addInPlace(Ecoeff)
                else:
                    newdiff[gen][imgen] = Ecoeff
            else:
                newdiff[gen] = {imgen : Ecoeff}
        else: # the image has a circle so there's two.
            # we act by split but it's uhh. complex.
            imgen2 = smoothdict[gen][1]
            cic = 0
            for i in range(len(circles)):
                if index in circles[i]:
                    cic = len(circles)-i-1
                    break
            coeff1 = Element(BasisElement(idem,imidem,circles,2**cic)) 
            if gen in newdiff:
                extradict = {imgen:coeff1, imgen2:Ecoeff}
                for igen, coef in extradict.items():
                    if igen in newdiff[gen]:
                        newdiff[gen][igen].addInPlace(coef)
                    else:
                        newdiff[gen][igen] = coef
            else:
                newdiff[gen] = extradict

    return newgens, newdiff,  newgr
    
def closed_braid(braid_word,  num_strands=False,  simplify=True, closed=True):
    """
    Returns the ChainCx corresponding to the braid closure of the specified braid word.
    In braid_word, i>0 stands for sigma_i, i<0 stands for sigma_i^-1.
    If num_strands isn't specified, deduces the number of strands from the highest braid generator appearing.
    Passes optional argument simplify to ChainCx.
    """
    if not num_strands:
        num_strands=max([abs(x) for x in braid_word])+1
    cap_tup = tuple((2*num_strands-i-1 for i in range(2*num_strands)))
    braid_terms = [cap_tup, ]
    for b in braid_word:
        if b>0:
            braid_terms.append(('d', b-1, 2*num_strands))
        else:
            braid_terms.append(('u', -b-1, 2*num_strands))
    braid_terms.append(cap_tup)
    if closed:
        return ChainCx(braid_terms,  simplify=simplify)
    return ChainCx(braid_terms[1:], simplify=simplify)
    
def glueCap(index, generators, differential,  grading):
    """ attaches a cap beginning at specified index
    """
    newgens = {} # new generator dictionary (gens:idems)
    gendict = {} # keys: old gens, values: new gens coming from them
    idemdict = {} # keys: idems, values: gens w that idem
    newgrs = {} #keys: new generators, values (h,q) bigradings

    # create new generators!
    for gen, idem in generators.items():
        newidem, extra = attachcap(index, idem)
        if extra:
            ngens = ['fa'+gen,'fx'+gen]
            newgrs['fa'+gen] = grading[gen]
            newgrs['fx'+gen] = (grading[gen][0], grading[gen][1]+2) #RL. Totally untested: this function never used anywhere.
        else:
            ngens = ['f'+gen]
            newgrs['f'+gen] = grading[gen]
        for ngen in ngens:
            newgens[ngen] = newidem

        gendict[gen] = ngens
        if newidem in idemdict:
            idemdict[newidem] += ngens.copy()
        else:
            idemdict[newidem] = ngens.copy()
    
    newdiff = {}
    # carry forward old differential.
    for oldgen, diff in differential.items():
        for gen in gendict[oldgen]:
            newdiff[gen] = {}
            if 'x' not in gen and 'a' not in gen:
                # no circles in gen!
                for diffgen in diff:
                    if len(gendict[diffgen]) == 1:
                        # no circle in the new gen
                        newelt = Element(dictionary={})
                        oldelt = diff[diffgen]
                        xlm1 = newgens[gen] # idem of gen
                        for basiselt in oldelt.dict:
                            xlm2 = attachcap(index,basiselt.cm2)[0]
                            newelt = newelt.add(Element(BasisElement(
                                xlm1,xlm2,loops(xlm1,xlm2)), oldelt.dict[basiselt]))
                        newdiff[gen]['f'+diffgen] = newelt
                        # are labels going to be fine???
                        # or do we need. uhh refactor??
                    else: #the new gen has a circle!
                        newelt1 = Element(dictionary={})
                        newelt2 = Element(dictionary={})
                        oldelt = diff[diffgen]
                        xlm1 = newgens[gen]
                        for basiselt in oldelt.dict:
                            # i SUSPECT that this actually won't work 
                            # on account of. refactor_split doesn't like me.
                            # actually just. the weirdness?? of cap.
                            nbe1, nbe2 = refactor_split(index, basiselt, xlm1)
                            if nbe1 != 0: # if it's zero there is no contribution.
                                newelt1.addInPlace(Element(nbe1, oldelt.dict[basiselt]))
                            newelt2.addInPlace(Element(nbe2, oldelt.dict[basiselt]))
                        newdiff[gen][unox + 'a' + diffgen] = newelt1
                        newdiff[gen][unox + 'x' + diffgen] = newelt2
            else: # gen has a circle
                for diffgen in diff:
                    if len(gendict[diffgen]) == 2:
                        # newgens have circles
                        newelt = Element(dictionary = {})
                        oldelt = diff[diffgen]
                        xlm1 = newgens[gen] #idem of generator we are finding diff of.
                        for basiselt in oldelt.dict:
                            xlm2 = attachcap(index,basiselt.cm2)[0] 
                            newelt.addInPlace(Element( 
                                BasisElement(xlm1,xlm2,loops(xlm1,xlm2)), oldelt.dict[basiselt]))
                        if newelt != Element():
                            newdiff[gen]['f'+diffgen] = newelt
                        else:
                            pass
                    else:
                        # merges!
                        extraX = 'x' in gen
                        newelt = Element(dictionary = {})
                        oldelt = diff[diffgen]
                        xlm1 = newgens[gen]
                        for basiselt in oldelt.dict:
                            # we need a refactor cap actually.
                            # the missing part!!
                            newdiffgen, xlm2, label = refactor(('cap', index, len(xlm1)), basiselt, extraX)
                            newelt = newelt.add(Element(newdiffgen, oldelt.dict[basiselt]))
                        newdiff[gen]['f'+diffgen] = newelt

    # no new differential!!

    return newgens, newdiff,  newgrs

def glueCup(index, generators, differential,  grading):
    """ attaches a cup beginning at specified index
    """
    newgens = {} # new generator dictionary (gens:idems)
    gendict = {} # keys: old gens, values: new gens coming from them
    idemdict = {} # keys: idems, values: gens w that idem
    newgrs = {} #New grading information

    # create new generators!
    for gen, idem in generators.items():
        ngen = 'f' + gen
        newidem = attachcup(index, idem)
        newgens[ngen] = newidem
        gendict[gen] = ngen
        newgrs[ngen] = grading[gen]
        if newidem in idemdict:
            idemdict[newidem] += [ngen]
        else:
            idemdict[newidem] = [ngen]
    
    newdiff = {}
    # carry forward old differential 
    for gen, diff in differential.items():
        ngen = gendict[gen]
        newdiff[ngen] = {}
        xlm1 = newgens[ngen]
        for diffgen, coeff in diff.items():
            newelt = Element(dictionary = {})
            for basiselt in coeff.dict:
                xlm2 = attachcup(index,basiselt.cm2)
                circles = loops(xlm1,xlm2)
                newlabels = basiselt.labels
                for i in range(len(circles)):
                    if index in circles[i]:
                        newlabels = newlabels[:i] + (0,) \
                            + newlabels[i:]
                        break
                newelt = newelt.add(Element(BasisElement(xlm1, xlm2,
                    circles, inlabels=newlabels), coeff.dict[basiselt]))
            newdiff[ngen]['f'+diffgen] = newelt
    
    # no new differential!!

    return newgens, newdiff,  newgrs


def _base_2_size(n):
    "Represent n base 2, then count the number of 1s in it."
    answer = 0
    m = n
    while m:
        answer += m%2
        m = int(m/2)
    return answer

def print_homol(x):
    "input is a dictionary with keys bigradings and values dimensions of homology. Prints the dictionary as a table."
    minh = min([m for (m,n) in x.keys()])
    maxh = max([m for (m,n) in x.keys()])
    minq = min([n for (m,n) in x.keys()])
    maxq = max([n for (m,n) in x.keys()])
    width = maxh-minh+1
    height = maxq-minq+1
    answer = list()
    answer.append(['q\h']+list(range(minh,maxh+1)))
    qrange = list(range(minq,maxq+1,2))
    qrange.reverse()
    for q in qrange:
        row = list()
        for i in range(minh,maxh+1):
            if (i,q) in x.keys():
                if x[(i,q)] > 0:
                    row.append(x[(i,q)])
                else:
                    row.append(' ')
            else:
                row.append(' ')
        answer.append(row)
    row_format = "{:>5}" * (width+1)
    for (i,row) in enumerate(answer):
        if i>0:
            print(row_format.format(repr(maxq-2*i+2),*row))
        else:
            print(row_format.format(*row))
    #return answer

class ModuleGenerator():
    """A generator of the module associated to a tangle."""
    def __init__(self, name, matching):
        "Name: a string that is the name of this generator. Matching: the crossingless matching this generator corresponds to."
        self.name = name
        self.matching = matching

    def toModuleElement(self):
        "Returns 1\otimes self, as a ModuleElement"
        idem = idempotent(self.matching)
        return ModuleElement({self:idem})

    def __rmul__(self, other):
        "Other should be an Element."
        return self.toModuleElement().__rmul__(other)

    
class ModuleElement(dict):
    """An element of the module associated to a tangle. 
    This is a formal linear combination of elements of the arc algebra (class Element) tensored with ModuleGenerators.
    Stored as a dictionary with keys ModuleGenerators and values Elements.
    """
    def __add__(self, other):
        "Other should also be a ModuleElement, or a dictionary with equivalent information."
        assert (isinstance(other, ModuleElement) or isinstance(other, dict))
        answer = dict()
        for x in other.keys():
            if x in self.keys():
                answer[x] = self[x]+other[x]
            else:
                answer[x] = other[x]
        for x in self.keys():
            if not (x in other.keys()):
                answer[x] = self[x]
        return ModuleElement(answer)

    def __rmul__(self, other):
        "Other should be an Element."
        answer = ME0
        for g in self.keys():
            answer += ModuleElement({g:other * self[g]})
        return answer

    def __eq__(self, other):
        selfd = dict()
        othd = dict()
        for x in self.keys():
            if self[x] != E0:
                selfd[x] = self[x]
        for x in other.keys():
            if other[x] != E0:
                othd[x] = other[x]
        return selfd == othd
    
#The module element zero
ME0 = ModuleElement({})
